#region modules
import logging 
#endregion

#region variables
log_mode = logging.INFO
logging.basicConfig(level=log_mode)
logger = logging.getLogger('xctph')
#endregion

#region functions
def get_logger():
    return logger
#endregion

#region classes
#endregion


#!/usr/bin/env python

import numpy as np
import scipy.io
import h5py
import xmltodict
import glob

from xctph.kpoints import get_all_kq_maps
from xctph.units import har2ryd, amu2me


def main(dirname_epw, prefix, nq, nc, nv, num_val, npool, switch_nu_and_cart=False, fname_out=None):
    """Routine to parse electron-phonon related files.

    Parameters
    ----------
    fname_nscf_xml : str
        Path to prefix.xml file writing during a nscf calculation in QE.
    nq : int
        Number of q-points on the 
    nv : int
        Number of valence states.
    nc : int
        NUmber of conduction states.
    num_val: int
        NUmber of valence bands in the system. 

    """

    def to_array(string):
        return np.genfromtxt(string.split())

    with open(dirname_epw + '/' + prefix + '.xml') as foo:
        data = xmltodict.parse(foo.read())

        # Determine number of pools calculation was run on.
        # npool = int(data[u'qes:espresso']['parallel_info']['npool'])  # Can get it wrong sometimes, so fixing with manual input. 
        prefix = data[u'qes:espresso']['input']['control_variables']['prefix']

        # Begin parsing structural information from input.
        f = data[u'qes:espresso']['input']

        nat = int(f['atomic_structure']['@nat'])
        alat = float(f['atomic_structure']['@alat'])
        ntyp = int(f['atomic_species']['@ntyp'])

        lat = np.zeros((3, 3), 'f8')
        pos = np.zeros((nat, 3), 'f8')
        species = dict()
        typlist = list()
        masses = np.zeros(ntyp, 'f8')

        f_cell = f['atomic_structure']['cell']
        for i in range(3):
            lat[i, :] = to_array(f_cell['a{}'.format(i + 1)])

        # Also get recip lat in atomic units (a.u.)! To get cartesian coords:
        # k^{cart} = np.dot(recip_lat, k^{crys}).
        # We use this later to convert derivatives, then:
        # dE^{cart} = np.dot(inv(recip_lat.T), dE^{crys})
        recip_lat = 2 * np.pi * np.linalg.inv(lat)

        if ntyp == 1:
            species[str(f['atomic_species']['species']['@name'])] = 0
            masses[0] = float(f['atomic_species']['species']['mass'])
        else:
            f_species = f['atomic_species']['species']
            for i in range(ntyp):
                species[str(f_species[i]['@name'])] = i
                masses[i] = float(f_species[i]['mass'])

        f_atom = f['atomic_structure']['atomic_positions']['atom']
        if nat == 1:
            typlist.append(f_atom['@name'])
            pos[0, :] = to_array(f_atom['#text'])
        else:
            for i in range(nat):
                typlist.append(f_atom[i]['@name'])
                pos[i, :] = to_array(f_atom[i]['#text'])

        typlist_num = [species[typ] for typ in typlist]

        # Begin parsing output.
        f = data[u'qes:espresso']['output']['band_structure']

        # if f['lsda'] == 2:
        #     raise NotImplemented('Spinor capability not yet implemented.')

        nbnd = int(f['nbnd'])
        # nocc = int(float(f['nelec']) / 2.0)
        nocc = int(num_val)
        nk = int(f['nks'])

        # Read in kpoints and energies.
        f = f['ks_energies']

        kpts_cart = np.zeros((nk, 3), 'f8')
        energies = np.zeros((nbnd, nk), 'f8')

        for ik in range(nk):
            kpts_cart[ik, :] = to_array(f[ik]['k_point']['#text'])
            energies[:, ik] = to_array(f[ik]['eigenvalues']['#text']) * har2ryd

    def cart2crys(kk):
        return np.dot(kk, (lat / alat).T)

    kpts = cart2crys(kpts_cart)

    nmodes = 3 * nat
    nk_per_pool = int(nk / npool)

    # Now confirm that the choice of nbnd begin/end makes sense
    nbnd_begin = nocc - nv
    nbnd_end = nocc + nc
    nbnd_red = nc + nv

    if nbnd_begin > nbnd_end:
        raise Exception('Keyword "nbnd_begin" cannot be greater than "nbnd_end".')
    elif nbnd_end > nbnd:
        raise Exception('Keyword "nbnd_begin" cannot be greater than "nbnd in QE".')
    if nbnd_begin < 0:
        raise Exception('Keyword "nbnd_begin" cannot be less than 0.')

    # Arrays to be used later
    gqk_ia = np.zeros((nbnd_red, nbnd_red, nk, nmodes, nq), 'c16')
    dynmat = np.zeros((nmodes, nmodes, nq), 'c16')

    # Loop through and read epb files
    ikk = 0
    for ipool in range(npool):

        print('Reading in {}.epb{}'.format(prefix, ipool + 1))

        f_in = '{}/{}.epb{}'.format(dirname_epw, prefix, ipool + 1)
        f = scipy.io.FortranFile(f_in, mode='r')

        # Define shape of all quantities to be read in.
        nqc_dim = 'i4'
        xqc_dim = '{}f8'.format(3 * nq)
        et_dim = '{}f8'.format(nbnd * nk_per_pool)

        dynq_dim = '{}c16'.format(nmodes * nmodes * nq)
        epmatq_dim = '{}c16'.format(nbnd_red * nbnd_red * nmodes * nk_per_pool * nq)
        zstar_dim = '{}f8'.format(3 * nmodes)
        epsi_dim = '{}f8'.format(3 * 3)


        # Read in all epb file
        record = f.read_record(nqc_dim, xqc_dim, et_dim, dynq_dim, epmatq_dim, zstar_dim, epsi_dim)

        if ipool == 0:
            nq_readin = record[0][0]
            qpts_cart = record[1].reshape(3, nq, order='F').T
            ekq = record[2].reshape(nbnd, nk_per_pool, order='F')
            dynmat = record[3].reshape(nmodes, nmodes, nq, order='F')
            qpts = cart2crys(qpts_cart)
            assert nq_readin == nq

        gqk_readin = record[4].reshape(nbnd_red, nbnd_red, nk_per_pool, nmodes, nq, order='F')

        # Trim down gqk_atom
        #print('WARNING: excluding bands not included in W90, this is how EPW with qe6.3 works.')
        gqk_ia[:, :, ikk:ikk + nk_per_pool, :, :] = gqk_readin[:, :, ...]

        ikk = (ipool + 1) * nk_per_pool


    # Rotate epb matrix elements from atomic to mode basis
    gkq_nu = np.zeros((nbnd_red, nbnd_red, nk, nmodes, nq), 'c16')
    frequencies = np.zeros((nmodes, nq), 'f8')

    print('Begin rotating from gkq from atomic to mode basis')

    for iq in range(nq):
        print('start iq={}'.format(iq))

        dyn = dynmat[..., iq]
        dynm = np.zeros((nmodes, nmodes), 'c16')

        # Compute mass reduced dyn (which we call dynm)
        for i in range(nat):
            for j in range(nat):
                m_i = masses[typlist_num[i]]
                m_j = masses[typlist_num[j]]

                fac = 1.0 / (np.sqrt(m_i * m_j) * amu2me / 2.)

                dynm[3 * i:3 * (i + 1), 3 * j:3 * (j + 1)] = dyn[3 * i:3 * (i + 1), 3 * j:3 * (j + 1)] * fac

        # Symmetrize dynm (which we call dynm_sym)
        dynm_sym = np.asarray((dynm + np.conj(dynm).T) / 2, 'c16')

        # Diagonalize dynm_sym
        w2, v = np.linalg.eigh(dynm_sym)

        # Store eigenfrequencies
        for imode in range(nmodes):

            if iq == 0 and imode < 3:
                frequencies[imode, iq] = 0.

            else:
                if w2[imode] < 0.0:
                    print('Warning: The frequency associated with mode ' \
                          'iq = {}, imode = {} is negative w ={}.' \
                          .format(iq, imode, w2[imode]) )

                frequencies[imode, iq] = np.sqrt(np.abs(w2[imode]))

        # Compute mass reduced normal modes
        vm = np.empty((3 * nat, nmodes), 'c16')

        for i in range(nat):
            m_i = masses[typlist_num[i]]
            vm[3 * i:3 * (i + 1), :] = \
                v[3 * i:3 * (i + 1), :] / np.sqrt(m_i * amu2me / 2.)

        # Loop through gkq_ia and transform into the mode basis
        gk_ai = gqk_ia[..., iq]
        gk_nu = np.zeros((nbnd_red, nbnd_red, nk, nmodes), 'c16')

        for ib in range(nbnd_red):
            for jb in range(nbnd_red):
                for ik in range(nk):
                    for imode in range(nmodes):
                        gk_nu[ib, jb, ik, imode] = np.dot(gk_ai[ib, jb, ik, :], vm[:, imode])

        for imode in range(nmodes):

            if iq == 0 and imode < 3:
                gkq_nu[..., imode, iq] = 0.

            else:
                gkq_nu[..., imode, iq] = gk_nu[..., imode] / np.sqrt(2 * frequencies[imode, iq])

    # We also compute the kq_maps in this
    k_plus_q_map = get_all_kq_maps(kpts, qpts)

    # Routines to write elph to hdf5 format
    if fname_out is None:
        fname_out = 'eph.h5'

    f = h5py.File(fname_out, 'w')

    f.create_group('gkq_header')
    f.create_group('gkq_data')

    sorting_dict = {

        # Header information.
        'gkq_header/ns': 1,
        'gkq_header/nbndskip': nbnd_begin,
        'gkq_header/nbnd': nbnd_red,
        'gkq_header/nocc': nocc,
        'gkq_header/nmode': nmodes,
        'gkq_header/nk': nk,
        'gkq_header/nq': nq,
        'gkq_header/kpts': kpts,
        'gkq_header/qpts': qpts,
        'gkq_header/recip_lat': recip_lat,


        # Information on k+q mappings.
        'gkq_mappings/k_plus_q_map': k_plus_q_map,

        # Data sets.
        'gkq_data/energies': energies[nbnd_begin:nbnd_end],
        'gkq_data/frequencies': frequencies,
        'gkq_data/g_nu': gkq_nu,
        'gkq_data/g_cart': gqk_ia,

    }

    if switch_nu_and_cart:
        sorting_dict['gkq_data/g_nu'] = gqk_ia
        sorting_dict['gkq_data/g_cart'] = gkq_nu

    for name, data in sorting_dict.items():
        f.create_dataset(name, data=data)

    f.close()

    return


if __name__ == '__main__':

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('dirname_epw')
    parser.add_argument('prefix')
    parser.add_argument('nq', type=int)
    parser.add_argument('nc', type=int)
    parser.add_argument('nv', type=int)
    parser.add_argument('num_val', type=int)
    parser.add_argument('npool', type=int)
    parser.add_argument('--switch_nu_and_cart', action='store_true', help='Hack to get excited state forces.')
    args = parser.parse_args()

    main(**vars(args))



#region modules
import numpy as np 
import glob 
import h5py 
from ase.units import Ry, eV
#endregion

#region variables
#endregion

#region functions
#endregion

#region classes
class Xct:
    def __init__(self, xct_dir_glob: str = './bseq_for_xctph/Q_*/eigenvectors.h5'):
        self.xct_dir_glob: str = xct_dir_glob  

    def read(self):
        # search for eigenvectors
        fnames = np.sort(glob.glob(self.xct_dir_glob))
        self.nQ = len(fnames)

        # allocate space for Qpts
        self.exciton_Q_shift = np.zeros((self.nQ, 3), 'f8')

        for iQ, fname in enumerate(fnames):

            # Reads in exciton header associated with current q-point.
            with h5py.File(fname, 'r') as f:

                # if iQ=0, store header information.
                if iQ == 0:

                    self.exciton_Q_shift[iQ, :] = f['exciton_header/kpoints/exciton_Q_shifts'][()]
                    self.nevecs = f['exciton_header/params/nevecs'][()]
                    self.ns = f['exciton_header/params/ns'][()]
                    self.nv = f['exciton_header/params/nv'][()]
                    self.nc = f['exciton_header/params/nc'][()]
                    self.nk = f['exciton_header/kpoints/nk'][()]
                    self.kpts = f['exciton_header/kpoints/kpts'][()]

                    # some other info needed
                    self.blat = f['mf_header/crystal/blat'][()]
                    self.bvec = f['mf_header/crystal/bvec'][()]

                    eigenvalues = np.zeros((self.nevecs, self.nQ), 'f8')
                    eigenvectors = np.zeros((2, self.ns, self.nv, self.nc, self.nk, self.nevecs, self.nQ), 'f8')

                # otherwise, check values 
                else:

                    self.exciton_Q_shift[iQ, :] = f['exciton_header/kpoints/exciton_Q_shifts'][()]
                    assert self.nevecs == f['exciton_header/params/nevecs'][()]
                    assert self.ns == f['exciton_header/params/ns'][()]
                    assert self.nv == f['exciton_header/params/nv'][()]
                    assert self.nc == f['exciton_header/params/nc'][()]
                    assert self.nk == f['exciton_header/kpoints/nk'][()]


                # reads in energies and exciton coefficients
                eigenvalues[:, iQ] = f['exciton_data/eigenvalues'][:self.nevecs]
                eigenvectors[..., iQ] = f['exciton_data/eigenvectors'][0,:self.nevecs,...].T

        self.eigs = eigenvalues / (Ry / eV)
        self.evecs = eigenvectors[0,...] + 1j * eigenvectors[1,...]

    def write(self):
        ex_data = {
            '/exciton_header/nevecs' : self.nevecs,
            '/exciton_header/ns' : self.ns,
            '/exciton_header/nv' : self.nv,
            '/exciton_header/nc' : self.nc,
            '/exciton_header/nk' : self.nk,
            '/exciton_header/nQ' : self.nQ,
            '/exciton_header/kpts' : self.kpts,
            '/exciton_header/center_of_mass_Q' : -1.0 * self.exciton_Q_shift,

            '/exciton_data/eigenvalues' : self.eigs,
            '/exciton_data/eigenvectors' : self.evecs,

            '/mf_header/crystal/blat' : self.blat,
            '/mf_header/crystal/bvec' : self.bvec,
        }
        with h5py.File('xct.h5', 'w') as f:
            f.create_group('exciton_header')
            f.create_group('exciton_data')
            for name, data in ex_data.items():
                f.create_dataset(name, data=data)

#endregion



#!/usr/bin/env python

import numpy as np
import h5py
from mpi4py import MPI
from xctph.kpoints import get_all_kq_maps


# TODO: Generalize to an shape. Currently only supports 2D shapes.
class ParSize:
    def __init__(self, shape, comm: MPI.Comm):
        self.shape = shape
        self.array_size: int = int(np.prod(np.array(shape)))
        self.comm: MPI.Comm = comm

        # Do the setup here.
        self.mpi_size = comm.Get_size()
        self.mpi_rank = comm.Get_rank()
        self._calc_local_ranges()

    def _calc_local_ranges(self):
       self.local_size_avg = self.array_size // self.mpi_size
       self.local_size_max = self.local_size_avg + self.array_size -  (self.array_size // self.mpi_size ) * self.mpi_size
       self.local_size = self.local_size_avg if self.mpi_rank != self.mpi_size-1 else self.local_size_max
       self.local_start = self.local_size_avg * self.mpi_rank  
       self.local_end = self.local_start + self.local_size 

    def get_local_range(self):
        return (self.local_start, self.local_end)

    def get_idx(self, linear_idx):
        return (linear_idx // self.shape[1], linear_idx % self.shape[0]) 

def compute_xctph(fname_eph_h5, fname_xct_h5, nbnd_xct, add_electron_part, add_hole_part):
    """ Compute the exciton-phonon matrix elements. """

    with h5py.File(fname_eph_h5, 'r') as f:
        nmodes = f['gkq_header/nmode'][()]
        nk_elph = f['gkq_header/nk'][()]
        kpts_elph = f['gkq_header/kpts'][()]
        nq = f['gkq_header/nq'][()]
        qpts = f['gkq_header/qpts'][()]
        k_plus_q_map = f['gkq_mappings/k_plus_q_map'][()]
        frequencies = f['gkq_data/frequencies'][()]
        gkq = f['gkq_data/g_nu'][()]

    with h5py.File(fname_xct_h5, 'r') as f:
        nbnd = f['/exciton_header/nevecs'][()]
        nv = f['/exciton_header/nv'][()]
        nc = f['/exciton_header/nc'][()]
        nk = f['/exciton_header/nk'][()]
        kpts = f['/exciton_header/kpts'][()]
        nQ = f['/exciton_header/nQ'][()]
        Qpts = f['/exciton_header/center_of_mass_Q'][()]
        energies = f['/exciton_data/eigenvalues'][()]
        avck = f['/exciton_data/eigenvectors'][()]


    # consistency checks:
    assert nk_elph == nk
    assert nbnd >= nbnd_xct

    # generate additional kq maps
    Q_plus_q_map = get_all_kq_maps(Qpts, qpts)
    k_minus_Q_map = get_all_kq_maps(kpts, Qpts, -1.0)

    # xct-ph matrix elements are packaged the same way as el-ph
    # gQq = np.zeros((nbnd_xct, nbnd_xct, nQ, nmodes, nq), 'c16')
    # Parallel version.
    comm = MPI.COMM_WORLD
    xctbnd_parsize = ParSize(shape=(nbnd_xct, nbnd_xct), comm=comm)
    xct_start, xct_end = xctbnd_parsize.get_local_range()
    gQq: np.ndarray = np.zeros((xct_end - xct_start, nQ, nmodes, nq), 'c16')

    cb = slice(nv, nv + nc)
    vb = slice(0, nv)

    for iQ in range(nQ):
      for iq in range(nq):
        iQ_plus_q = Q_plus_q_map[iQ, iq]

        for ik in range(nk):
          ik_plus_q = k_plus_q_map[ik, iq]
          ik_minus_Q = k_minus_Q_map[ik, iQ]

        #   for mb in range(nbnd_xct):
        #     for nb in range(nbnd_xct):

        #         # electron channel
        #         aQ_e = avck[0, :, :, ik, nb, iQ]
        #         aQq_e = avck[0, :, :, ik_plus_q, mb, iQ_plus_q]
        #         gkq_e = gkq[cb, cb, ik, :, iq]

        #         if add_electron_part:
        #             gQq[mb, nb, iQ, :, iq] += np.einsum('vc,cdn,vd->n', aQq_e.conj(), gkq_e, aQ_e)

        #         # hole channel
        #         aQ_h = avck[0, :, :, ik_plus_q, nb, iQ]
        #         aQq_h = avck[0, :, :, ik_plus_q, mb, iQ_plus_q]
        #         gkq_h = gkq[vb, vb, ik_minus_Q, :, iq][::-1, ::-1, :]

        #         if add_hole_part:
        #             gQq[mb, nb, iQ, :, iq] -= np.einsum('vc,wvn,wc->n', aQq_h.conj(), gkq_h, aQ_h)

        # Parallel version. 
          for mnb in range(xct_start, xct_end):
            mb, nb = xctbnd_parsize.get_idx(mnb)
            # electron channel
            aQ_e = avck[0, :, :, ik, nb, iQ]
            aQq_e = avck[0, :, :, ik_plus_q, mb, iQ_plus_q]
            gkq_e = gkq[cb, cb, ik, :, iq]

            if add_electron_part:
                gQq[mnb - xct_start, iQ, :, iq] += np.einsum('vc,cdn,vd->n', aQq_e.conj(), gkq_e, aQ_e)

            # hole channel
            aQ_h = avck[0, :, :, ik_plus_q, nb, iQ]
            aQq_h = avck[0, :, :, ik_plus_q, mb, iQ_plus_q]
            gkq_h = gkq[vb, vb, ik_minus_Q, :, iq][::-1, ::-1, :]

            if add_hole_part:
                gQq[mnb - xct_start, iQ, :, iq] -= np.einsum('vc,wvn,wc->n', aQq_h.conj(), gkq_h, aQ_h)



    xctph_dict = {
        # header information.
        'ns': 1,
        'nbndskip': 0,
        'nbnd': nbnd_xct,
        'nocc': 0,
        'nmode': nmodes,
        'nQ': nQ,
        'nq': nq,
        'Qpts': Qpts,
        'qpts': qpts,

        # Q+q mappings.
        'Q_plus_q_map': Q_plus_q_map,

        # energies, frequencies, and matrix elements.
        'energies': energies[:nbnd_xct, :],
        'frequencies': frequencies,
        'xctph' : gQq,

    }

    # f = h5py.File('xctph.h5', 'w')
    # for name, data in xctph_dict.items():
    #     f.create_dataset(name, data=data)
    # f.close()

    # Parallel version. 
    with h5py.File('xctph.h5', 'w', driver='mpio', comm=comm) as f:
        # Write everything except the xctph data array, as that needs to be written in parallel.
        for name, data in xctph_dict.items():
            if name != 'xctph':
                f.create_dataset(name, data=data)
            else:
                # Write the xctph array in parallel.
                ds_xctph_linear = f.create_dataset('xctph_linear', shape=(nbnd_xct*nbnd_xct, nQ, nmodes, nq), dtype=gQq.dtype)
                ds_xctph_linear[xct_start:xct_end, ...] = data

    # # Reshape only on main node.
    if xctbnd_parsize.mpi_rank==0:
        with h5py.File('xctph.h5', 'a') as w:
            xctph = w['xctph_linear'][:].reshape(nbnd_xct, nbnd_xct, nQ, nmodes, nq)
            del w['xctph_linear']
            w.create_dataset('xctph', data=xctph)

    comm.Barrier()

    return


if __name__ == '__main__':

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('fname_eph_h5')
    parser.add_argument('fname_xct_h5')
    parser.add_argument('nbnd_xct', type=int)
    parser.add_argument('--add_electron_part', action='store_true', help='Add electron part to xctph computation')
    parser.add_argument('--add_hole_part', action='store_true', help='Add hole part to xctph computation')
    args = parser.parse_args()

    compute_xctph(**vars(args))



#region modules
import os 
import h5py 
import xmltodict
import numpy as np 
from ase.units import Hartree, Rydberg, _amu, _me
import scipy 
from xctph.utils.k_plus_q import get_all_kq_maps
import jmespath
from xctph.utils.logging import get_logger
#endregion

#region variables
logger = get_logger()
#endregion

#region functions
#endregion

#region classes
class Elph:
    def __init__(
        self,
        nocc: int,
        nc: int,
        nv: int,
        npool: int = 1,
        elph_dirname: str = './save',
        qexml_filename: str = 'wfn.xml',
    ):
        self.elph_dirname: str = elph_dirname
        self.qexml_filename: str = qexml_filename
        self.npool: int = npool
        self.nocc: int = nocc
        self.nc: int = nc
        self.nv: int = nv 

        # Other vars.
        self.param_dict: dict = None        # Combination of qexml data and input_dict data elements that are needed. 

    def _to_array(self, string):
        return np.genfromtxt(string.split())
    
    def _cart2crys(self, kk, lat, alat):
        return np.dot(kk, (lat / alat).T)

    def read_from_input_and_qe(self):
        with open(os.path.join(self.elph_dirname, self.qexml_filename), 'r')as read_stream:
            qexml = xmltodict.parse(read_stream.read())

        # Update. Order is from base file used for refactoring. 
        self.prefix = qexml[u'qes:espresso']['input']['control_variables']['prefix']
        
        # parse qexml input.
        input_node = qexml[u'qes:espresso']['input']
        self.nat = int(input_node['atomic_structure']['@nat'])
        self.alat = float(input_node['atomic_structure']['@alat'])
        self.ntyp = int(input_node['atomic_species']['@ntyp'])
        self.lat = np.zeros((3, 3), 'f8')
        self.pos = np.zeros((self.nat, 3), 'f8')
        self.species = dict()
        self.typlist = list()
        self.masses = np.zeros(self.ntyp, 'f8')
        for i in range(3):
            self.lat[i, :] = self._to_array(input_node['atomic_structure']['cell']['a{}'.format(i + 1)])
        self.recip_lat = 2 * np.pi * np.linalg.inv(self.lat)
        if self.ntyp == 1:
            self.species[str(input_node['atomic_species']['species']['@name'])] = 0
            self.masses[0] = float(input_node['atomic_species']['species']['mass'])
        else:
            for i in range(self.ntyp):
                self.species[str(input_node['atomic_species']['species'][i]['@name'])] = i
                self.masses[i] = float(input_node['atomic_species']['species'][i]['mass'])
        f_atom = input_node['atomic_structure']['atomic_positions']['atom']
        if self.nat == 1:
            self.typlist.append(f_atom['@name'])
            self.pos[0, :] = self._to_array(f_atom['#text'])
        else:
            for i in range(self.nat):
                self.typlist.append(f_atom[i]['@name'])
                self.pos[i, :] = self._to_array(f_atom[i]['#text'])

        self.typlist_num = [self.species[typ] for typ in self.typlist]

        # parse qexml output.
        output_node = qexml[u'qes:espresso']['output']['band_structure']
        self.nbnd = self.nocc + self.nc
        self.nbnd_end = self.nocc + self.nc 
        self.nbnd_begin = self.nocc - self.nv 
        self.nbnd_red = self.nc + self.nv 
        self.nk = int(output_node['nks'])
        self.nq = self.nk 
        self.kpts_cart = np.zeros((self.nk, 3), 'f8')
        for ik in range(self.nk):
            self.kpts_cart[ik, :] = self._to_array(output_node['ks_energies'][ik]['k_point']['#text'])
        self.kpts = self._cart2crys(self.kpts_cart, self.lat, self.alat)
        self.nmodes = 3 * self.nat 
        self.nk_per_pool = int(self.nk / self.npool)

        if self.nbnd_begin > self.nbnd_end:
            raise Exception('Keyword "nbnd_begin" cannot be greater than "nbnd_end".')
        elif self.nbnd_end > self.nbnd:
            raise Exception('Keyword "nbnd_begin" cannot be greater than "nbnd in QE".')
        if self.nbnd_begin < 0:
            raise Exception('Keyword "nbnd_begin" cannot be less than 0.')
        
    def read_from_epb(self):
        self.elph_cart = np.zeros((self.nbnd_red, self.nbnd_red, self.nk, self.nmodes, self.nq), 'c16')
        
        # Loop through and read epb files
        ikk = 0
        for ipool in range(self.npool):

            print('Reading in {}.epb{}'.format(self.prefix, ipool + 1))

            f_in = '{}/{}.epb{}'.format(self.elph_dirname, self.prefix, ipool + 1)
            f = scipy.io.FortranFile(f_in, mode='r')

            # Define shape of all quantities to be read in.
            nqc_dim = 'i4'
            xqc_dim = '{}f8'.format(3 * self.nq)
            et_dim = '{}f8'.format(self.nbnd * self.nk_per_pool)

            dynq_dim = '{}c16'.format(self.nmodes * self.nmodes * self.nq)
            epmatq_dim = '{}c16'.format(self.nbnd_red * self.nbnd_red * self.nmodes * self.nk_per_pool * self.nq)
            zstar_dim = '{}f8'.format(3 * self.nmodes)
            epsi_dim = '{}f8'.format(3 * 3)

            # Read in all epb file
            record = f.read_record(nqc_dim, xqc_dim, et_dim, dynq_dim, epmatq_dim, zstar_dim, epsi_dim)

            if ipool == 0:
                nq_readin = record[0][0]
                qpts_cart = record[1].reshape(3, self.nq, order='F').T
                ekq = record[2].reshape(self.nbnd, self.nk_per_pool, order='F')
                self.dynmat = record[3].reshape(self.nmodes, self.nmodes, self.nq, order='F')
                self.qpts = self._cart2crys(qpts_cart, self.lat, self.alat)
                assert nq_readin == self.nq

            gqk_readin = record[4].reshape(self.nbnd_red, self.nbnd_red, self.nk_per_pool, self.nmodes, self.nq, order='F')
            self.elph_cart[:, :, ikk:ikk + self.nk_per_pool, :, :] = gqk_readin[:, :, ...]

            ikk = (ipool + 1) * self.nk_per_pool

    def calc_dyn(self):
        pass

    def rotate_to_mode_basis(self):
        # Rotate epb matrix elements from atomic to mode basis
        self.elph_mode = np.zeros((self.nbnd_red, self.nbnd_red, self.nk, self.nmodes, self.nq), 'c16')
        self.frequencies = np.zeros((self.nmodes, self.nq), 'f8')
        self.phevecs = np.zeros((3 * self.nat, self.nmodes, self.nq), 'c16')
        self.phmasses = np.zeros((3 * self.nat), 'f8')
        self.phevecs_massred = np.zeros((3 * self.nat, self.nmodes, self.nq), 'c16') 

        print('Begin rotating from gkq from atomic to mode basis')

        for iq in range(self.nq):
            print('start iq={}'.format(iq))

            dyn = self.dynmat[..., iq]
            dynm = np.zeros((self.nmodes, self.nmodes), 'c16')

            # Compute mass reduced dyn (which we call dynm)
            for i in range(self.nat):
                for j in range(self.nat):
                    m_i = self.masses[self.typlist_num[i]]
                    m_j = self.masses[self.typlist_num[j]]

                    fac = 1.0 / (np.sqrt(m_i * m_j) * _amu/_me / 2.)

                    dynm[3 * i:3 * (i + 1), 3 * j:3 * (j + 1)] = dyn[3 * i:3 * (i + 1), 3 * j:3 * (j + 1)] * fac

            # Symmetrize dynm (which we call dynm_sym)
            dynm_sym = np.asarray((dynm + np.conj(dynm).T) / 2, 'c16')

            # frequencies.
            w2, self.phevecs[:, :, iq] = np.linalg.eigh(dynm_sym)
            for imode in range(self.nmodes):

                if iq == 0 and imode < 3:
                    self.frequencies[imode, iq] = 0.

                else:
                    if w2[imode] < 0.0:
                        print('Warning: The frequency associated with mode ' \
                            'iq = {}, imode = {} is negative w ={}.' \
                            .format(iq, imode, w2[imode]) )

                    self.frequencies[imode, iq] = np.sqrt(np.abs(w2[imode]))

            # modes. (mass reduced normal modes.)

            for i in range(self.nat):
                m_i = self.masses[self.typlist_num[i]]
                self.phevecs_massred[3 * i:3 * (i + 1), :, iq] = \
                    self.phevecs[3 * i:3 * (i + 1), :, iq] / np.sqrt(m_i * _amu/ _me / 2.)
                self.phmasses[3 * i: 3 * (i + 1)] = m_i * _amu / _me / 2.   # In Rydberg units.
                
            # Loop through gkq_ia and transform into the mode basis
            gk_ai = self.elph_cart[..., iq]
            gk_nu = np.zeros((self.nbnd_red, self.nbnd_red, self.nk, self.nmodes), 'c16')

            for ib in range(self.nbnd_red):
                for jb in range(self.nbnd_red):
                    for ik in range(self.nk):
                        for imode in range(self.nmodes):
                            gk_nu[ib, jb, ik, imode] = np.dot(gk_ai[ib, jb, ik, :], self.phevecs_massred[:, imode, iq])

            for imode in range(self.nmodes):
                if iq == 0 and imode < 3:
                    self.elph_mode[..., imode, iq] = 0.
                else:
                    self.elph_mode[..., imode, iq] = gk_nu[..., imode] / np.sqrt(2 * self.frequencies[imode, iq])

    def calc_k_plus_q_map(self):
        # We also compute the kq_maps in this
        self.k_plus_q_map = get_all_kq_maps(self.kpts, self.qpts)
        self.k_minus_q_map = get_all_kq_maps(self.kpts, self.qpts, plus_or_minus=-1.0)

    def read(self):
        self.read_from_input_and_qe()
        self.read_from_epb()
        self.calc_dyn()
        self.rotate_to_mode_basis()   
        self.calc_k_plus_q_map()

    def write(self):
        # Everything i
        sorting_dict = {
            # Header information.
            'elph_header/ns': 1,
            'elph_header/nbndskip': self.nbnd_begin,
            'elph_header/nbnd': self.nbnd_red,
            'elph_header/nocc': self.nocc,
            'elph_header/nmode': self.nmodes,
            'elph_header/nk': self.nk,
            'elph_header/nq': self.nq,
            'elph_header/kpts': self.kpts,
            'elph_header/qpts': self.qpts,
            'elph_header/recip_lat': self.recip_lat,
            
            # Information on k+q mappings.
            'elph_header/k_plus_q_map': self.k_plus_q_map,
            'elph_header/k_minus_q_map': self.k_minus_q_map,

            # Data sets.
            'elph_data/frequencies': self.frequencies,
            'elph_data/phevecs': self.phevecs,
            'elph_data/phevecs_massred': self.phevecs_massred,
            'elph_data/phmasses': self.phmasses,
            'elph_data/elph_mode': self.elph_mode,
            'elph_data/elph_cart': self.elph_cart,
        }
        
        with h5py.File('elph.h5', 'w') as f:
            for name, data in sorting_dict.items():
                f.create_dataset(name, data=data)

#endregion



#region modules
import numpy as np
import copy
from scipy import spatial
#endregion

#region variables
#endregion

#region functions
def unit_range(klist_in):

    tol = 1e-6
    klist = copy.copy(klist_in)

    for ik, k in enumerate(klist):
        for i, kx in enumerate(k):

            while kx < -tol:
                kx = kx + 1.0
            while kx > 1.0 -tol:
                kx = kx - 1.0

            k[i] = kx

        klist[ik,:] = k

    return klist

def find_kpt(ktargets, klist):

    klist = unit_range(klist)
    ktargets = unit_range(ktargets)

    tree = spatial.KDTree(klist)
    ik_addr = list()
    for k in ktargets:
        d, i = tree.query(k)
        if d > 1e-6:
            print('kpt not found:', k)
            i = None

        ik_addr.append(i)

    return ik_addr

def get_all_kq_maps(kpts=None, qpts=None, plus_or_minus=1.0):

    nk = np.shape(kpts)[0]
    nq = np.shape(qpts)[0]
    kq_maps = np.zeros((nk, nq), dtype=int)

    for iq, q in enumerate(qpts):
        kqpts = np.array(kpts) + plus_or_minus * np.array(q)
        kq_maps[:, iq] = find_kpt(kqpts, kpts)
  
    return kq_maps
#endregion

#region classes
#endregion



#region modules
from xctph.elph import Elph
from xctph.xct import Xct
from xctph.utils.k_plus_q import get_all_kq_maps
from mpi4py import MPI 
import h5py
import numpy as np 
from ase.units import Bohr, Ha, Ry, eV, Ang 
from xctph.old_files.symmetrize_mtxel import sym_mtxel
import jmespath
#endregion

#region variables
#endregion

#region functions
#endregion

#region classes
class ParSize:
    def __init__(self, shape, comm: MPI.Comm):
        self.shape = shape
        self.array_size: int = int(np.prod(np.array(shape)))
        self.comm: MPI.Comm = comm

        # Do the setup here.
        self.mpi_size = comm.Get_size()
        self.mpi_rank = comm.Get_rank()
        self._calc_local_ranges()

    def _calc_local_ranges(self):
       self.local_size_avg = self.array_size // self.mpi_size
       self.local_size_max = self.local_size_avg + self.array_size -  (self.array_size // self.mpi_size ) * self.mpi_size
       self.local_size = self.local_size_avg if self.mpi_rank != self.mpi_size-1 else self.local_size_max
       self.local_start = self.local_size_avg * self.mpi_rank  
       self.local_end = self.local_start + self.local_size 

    def get_local_range(self):
        return (self.local_start, self.local_end)

    def get_idx(self, linear_idx):
        assert linear_idx < self.array_size, f'Linear idx: {linear_idx} is greater than array size: {self.array_size}'
        
        array_idx = list(None for _ in range(len(self.shape)))
        current_index = linear_idx
        for dim in range(len(self.shape)-1):
            stride = int(np.prod(self.shape[dim+1:]))
            array_idx[dim] = current_index // stride
            current_index = current_index % stride
        array_idx[-1] = current_index 

        return array_idx 

        return output 

class Xctph:
    def __init__(
        self, 
        nocc: int,
        nc: int,
        nv: int,
        npool: int = 1,
        nxct: int=10,
        add_electron_part: bool = True,
        add_hole_part: bool = True,
    ):
        self.add_electron_part = add_electron_part
        self.add_hole_part = add_hole_part
        self.nxct: int = nxct

        self.nocc: int = nocc
        self.nc: int = nc
        self.nv: int = nv
        self.npool: int = npool

    def generate_elph(self):
        elph = Elph(
            nocc=self.nocc,
            nc=self.nc,
            nv=self.nv,
            npool=self.npool,
        )
        elph.read()
        elph.write()

    def read_elph(self):
        with h5py.File('./elph.h5', 'r') as f:
            self.nmodes = f['elph_header/nmode'][()]
            self.nk_elph = f['elph_header/nk'][()]
            self.kpts_elph = f['elph_header/kpts'][()]
            self.nq = f['elph_header/nq'][()]
            self.qpts = f['elph_header/qpts'][()]
            self.k_plus_q_map = f['elph_header/k_plus_q_map'][()]
            self.k_minus_q_map = f['elph_header/k_minus_q_map'][()]
            self.frequencies = f['elph_data/frequencies'][()]
            self.gkq = f['elph_data/elph_mode'][()]

    def generate_xct(self):
        xct = Xct()
        xct.read()
        xct.write()

    def read_xct(self):
        with h5py.File('./xct.h5', 'r') as f:
            self.nbnd = f['/exciton_header/nevecs'][()]
            self.nv = f['/exciton_header/nv'][()]
            self.nc = f['/exciton_header/nc'][()]
            self.nk = f['/exciton_header/nk'][()]
            self.kpts = f['/exciton_header/kpts'][()]
            self.nQ = f['/exciton_header/nQ'][()]
            self.Qpts = f['/exciton_header/center_of_mass_Q'][()]
            self.energies = f['/exciton_data/eigenvalues'][()]
            self.avck = f['/exciton_data/eigenvectors'][()]

    def calc(self):
        # Preliminaries.
        self.comm = MPI.COMM_WORLD
        # Only generate from the rank 0. 
        if self.comm.Get_rank()==0:
            self.generate_elph()
            self.generate_xct()
        self.read_elph()
        self.read_xct()

        # Checks.
        assert self.nk_elph == self.nk
        assert self.nbnd >= self.nxct

        # Calculate.
        # generate additional kq maps
        self.Q_plus_q_map = get_all_kq_maps(self.Qpts, self.qpts)
        k_minus_Q_map = get_all_kq_maps(self.kpts, self.Qpts, -1.0)

        # Parallel version.
        self.xctbnd_parsize = ParSize(shape=(self.nxct, self.nxct), comm=self.comm)
        self.xct_start, self.xct_end = self.xctbnd_parsize.get_local_range()
        self.gQq_eh: np.ndarray = np.zeros((self.xct_end - self.xct_start, self.nQ, self.nmodes, self.nq), 'c16')
        self.gQq_e: np.ndarray = np.zeros((self.xct_end - self.xct_start, self.nQ, self.nmodes, self.nq), 'c16')
        self.gQq_h: np.ndarray = np.zeros((self.xct_end - self.xct_start, self.nQ, self.nmodes, self.nq), 'c16')

        cb = slice(self.nv, self.nv + self.nc)
        vb = slice(0, self.nv)

        for iQ in range(self.nQ):
            for iq in range(self.nq):
                iQ_plus_q = self.Q_plus_q_map[iQ, iq]

                for ik in range(self.nk):
                    ik_plus_q = self.k_plus_q_map[ik, iq]
                    ik_minus_Q = k_minus_Q_map[ik, iQ]

                    for mnb in range(self.xct_start, self.xct_end):
                        mb, nb = self.xctbnd_parsize.get_idx(mnb)
                        # electron channel
                        aQ_e = self.avck[0, :, :, ik, nb, iQ]
                        aQq_e = self.avck[0, :, :, ik_plus_q, mb, iQ_plus_q]
                        gkq_e = self.gkq[cb, cb, ik, :, iq]

                        if self.add_electron_part:
                            self.gQq_eh[mnb - self.xct_start, iQ, :, iq] += np.einsum('vc,cdn,vd->n', aQq_e.conj(), gkq_e, aQ_e)
                            self.gQq_e[mnb - self.xct_start, iQ, :, iq] += np.einsum('vc,cdn,vd->n', aQq_e.conj(), gkq_e, aQ_e)

                        # hole channel
                        aQ_h = self.avck[0, :, :, ik_plus_q, nb, iQ]
                        aQq_h = self.avck[0, :, :, ik_plus_q, mb, iQ_plus_q]
                        gkq_h = self.gkq[vb, vb, ik_minus_Q, :, iq][::-1, ::-1, :]

                        if self.add_hole_part:
                            self.gQq_eh[mnb - self.xct_start, iQ, :, iq] -= np.einsum('vc,wvn,wc->n', aQq_h.conj(), gkq_h, aQ_h)
                            self.gQq_h[mnb - self.xct_start, iQ, :, iq] -= np.einsum('vc,wvn,wc->n', aQq_h.conj(), gkq_h, aQ_h)

                    if self.comm.Get_rank()==0:
                        print(f'Done Q: {iQ}, q: {iq}, k: {ik}', flush=True)
        
        print('Done loop', flush=True)
        
    def write(self):
        xctph_dict = {
            # header information.
            'ns': 1,
            'nbndskip': 0,
            'nbnd': self.nxct,
            'nocc': 0,
            'nmode': self.nmodes,
            'nQ': self.nQ,
            'nq': self.nq,
            'Qpts': self.Qpts,
            'qpts': self.qpts,

            # Q+q mappings.
            'Q_plus_q_map': self.k_plus_q_map,
            'Q_minus_q_map': self.k_minus_q_map,

            # energies, frequencies, and matrix elements.
            'energies': self.energies[:self.nxct, :],
            'frequencies': self.frequencies,
            'xctph_eh' : self.gQq_eh,
            'xctph_e' : self.gQq_e,
            'xctph_h' : self.gQq_h,
        }

        # # Serial version:
        # with h5py.File('xctph.h5', 'w') as f:
        #     for name, data in xctph_dict.items():
        #         f.create_dataset(name, data=data)

        # Parallel version.
        with h5py.File('xctph.h5', 'w', driver='mpio', comm=self.comm) as f:
            # Write everything except the xctph data array, as that needs to be written in parallel.
            for name, data in xctph_dict.items():
                if 'xctph' in name:
                    # Write the xctph array in parallel.
                    ds_xctph_linear = f.create_dataset(f'{name}_linear', shape=(self.nxct*self.nxct, self.nQ, self.nmodes, self.nq), dtype=self.gQq_eh.dtype)
                    ds_xctph_linear[self.xct_start:self.xct_end, ...] = data

                    # # Create virtual dataset for reshape.
                    layout = h5py.VirtualLayout(shape=(self.nxct, self.nxct, self.nQ, self.nmodes, self.nq), dtype='c16')
                    source = h5py.VirtualSource('xctph.h5', f'{name}_linear', shape=(self.nxct*self.nxct, self.nQ, self.nmodes, self.nq), dtype='c16')

                    layout[:, :, :, :, :] = source[:, :, :, :]
                    ds_vx = f.create_virtual_dataset(f'{name}', layout)
                else:
                    f.create_dataset(name, data=data)


        # # # Reshape only on main node.
        # if self.xctbnd_parsize.mpi_rank==0:
        #     with h5py.File('xctph.h5', 'a') as w:
        #         xctph = w['xctph_linear'][:].reshape(self.nbnd_xct, self.nbnd_xct, self.nQ, self.nmodes, self.nq)
        #         del w['xctph_linear']
        #         w.create_dataset('xctph', data=xctph)

        self.comm.Barrier()

    def print_summary(self, nbnd: int = None):
        with h5py.File('xctph.h5', 'r') as f:
            # band information
            nbnd_read = f['nbnd'][()]
            if nbnd is None or nbnd > nbnd_read:
                nbnd = nbnd_read   
            nmode = f['nmode'][()]
        
            # k/q point information
            Qpts = f['Qpts'][()]
            qpts = f['qpts'][()]
            Q_plus_q_map = f['Q_plus_q_map'][()]
        
            # state information information
            energies = f['energies'][()]
            frequencies = f['frequencies'][()]
            gkq = f['xctph_eh'][:nbnd, :nbnd, ...]

    
        # prepare to write
        with open('xctph.dat', 'w') as f:

            for iQ, Q in enumerate(Qpts):
                for iq, q in enumerate(qpts):

                    # find index for the point at k+q
                    iQq = Q_plus_q_map[iQ, iq]

                    # write k, q point information
                    f.write('\n{:^20s}{Q[0]:8.4f}{Q[1]:8.4f}{Q[2]:8.4f}\n'
                            .format('kpt = ', Q=Q))

                    f.write('{:^20s}{q[0]:8.4f}{q[1]:8.4f}{q[2]:8.4f}\n\n'
                            .format('qpt = ', q=q))

                    f.write('{:>6s}{:>6s}{:>7s}{:>10s}{:>11s}{:>13s}{:>13s} '
                            '{:>12s}{:>11s}{:>14s}\n'
                            .format('nbnd', 'mbnd', 'imode',
                                    'eSQ (eV)', 'eS\'Qq (eV)', 'omega (meV)',
                                    'Re[G] (meV)', 'Im[G] (meV)', '|G| (meV)',
                                    'Sym|G| (meV)'
                                    )
                            )

                    eQ = energies[:, iQ] * Ry/eV
                    eQq = energies[:, iQq] * Ry/eV
                    freq = frequencies[:, iq] * Ry/eV * 1000
                    g =  gkq[:, :, iQ, :, iq] * Ry/eV * 1000
                    g_sym = sym_mtxel(g, eQ, eQq, freq)

                    for nb in range(nbnd):
                        for mb in range(nbnd):
                            for imode in range(nmode):
                            # Large write statement
                                f.write(
                                    '{:>6d}{:>6d}{:>7d}{:>10.3f}{:>11.3f}'
                                    '{:>13.2f}{:>13.2f}{:>12.2f}{:>11.2f}'
                                    '{:>14.2f}\n'
                                    .format(nb + 1,
                                            mb + 1,
                                            imode + 1,
                                            eQ[nb],
                                            eQq[mb],
                                            freq[imode],
                                            np.real(g[mb, nb, imode]),
                                            np.imag(g[mb, nb, imode]),
                                            np.abs(g[mb, nb, imode]),
                                            np.abs(g_sym[mb, nb, imode])
                                            )
                                )

#endregion


#!/usr/bin/env python

import numpy as np
import h5py
import glob

from xctph.units import ryd2ev


def parse_excitons(path_to_eigenvectors):
    """Routine to read in bse excitons computed on some q-grid.

    Parameters
    ----------
    eigenvector_fnames list(str)
        Paths to binary eigenvector files produced by BGW absorption
        calculation.
    Returns
    -------
    ex_data : dict
        Dictionary containing quantities related to exction calculation,
        including exciton energies and coefficients.
    """

    # search for eigenvectors
    fnames = np.sort(glob.glob(path_to_eigenvectors))
    nQ = len(fnames)


    # allocate space for Qpts
    exciton_Q_shift = np.zeros((nQ, 3), 'f8')

    for iQ, fname in enumerate(fnames):

      # Reads in exciton header associated with current q-point.
      with h5py.File(fname, 'r') as f:

         # if iQ=0, store header information.
         if iQ == 0:

             exciton_Q_shift[iQ, :] = f['exciton_header/kpoints/exciton_Q_shifts'][()]
             nevecs = f['exciton_header/params/nevecs'][()]
             ns = f['exciton_header/params/ns'][()]
             nv = f['exciton_header/params/nv'][()]
             nc = f['exciton_header/params/nc'][()]
             nk = f['exciton_header/kpoints/nk'][()]
             kpts = f['exciton_header/kpoints/kpts'][()]

             # some other info needed
             blat = f['mf_header/crystal/blat'][()]
             bvec = f['mf_header/crystal/bvec'][()]

             eigenvalues = np.zeros((nevecs, nQ), 'f8')
             eigenvectors = np.zeros((2, ns, nv, nc, nk, nevecs, nQ), 'f8')

         # otherwise, check values 
         else:

             exciton_Q_shift[iQ, :] = f['exciton_header/kpoints/exciton_Q_shifts'][()]
             assert nevecs == f['exciton_header/params/nevecs'][()]
             assert ns == f['exciton_header/params/ns'][()]
             assert nv == f['exciton_header/params/nv'][()]
             assert nc == f['exciton_header/params/nc'][()]
             assert nk == f['exciton_header/kpoints/nk'][()]


         # reads in energies and exciton coefficients
         eigenvalues[:, iQ] = f['exciton_data/eigenvalues'][:nevecs]
         eigenvectors[..., iQ] = f['exciton_data/eigenvectors'][0,:nevecs,...].T

    eigenvalues = eigenvalues / ryd2ev
    eigenvectors_cplx = eigenvectors[0,...] + 1j * eigenvectors[1,...]

    
    # write exciton to hdf5
    fname_out = 'xct.h5'
    f = h5py.File(fname_out, 'w')

    f.create_group('exciton_header')
    f.create_group('exciton_data')

    ex_data = {

        '/exciton_header/nevecs' : nevecs,
        '/exciton_header/ns' : ns,
        '/exciton_header/nv' : nv,
        '/exciton_header/nc' : nc,
        '/exciton_header/nk' : nk,
        '/exciton_header/nQ' : nQ,
        '/exciton_header/kpts' : kpts,
        '/exciton_header/center_of_mass_Q' : -1.0 * exciton_Q_shift,

        '/exciton_data/eigenvalues' : eigenvalues,
        '/exciton_data/eigenvectors' : eigenvectors_cplx,

        '/mf_header/crystal/blat' : blat,
        '/mf_header/crystal/bvec' : bvec,
    }

    for name, data in ex_data.items():
        f.create_dataset(name, data=data)

    f.close()

    return


if __name__ == '__main__':

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('path_to_eigenvectors')
    args = parser.parse_args()

    parse_excitons(**vars(args))



#!/usr/bin/env python

import h5py
import numpy as np

from xctph.units import ryd2ev, ryd2mev, bohr2ang
from xctph.symmetrize_mtxel import sym_mtxel


def print_xctph(fname_xctph_h5, switch_nu_and_cart_eVA=False):
  """Prints xctph matrix elements EPW style. """

  with h5py.File(fname_xctph_h5, 'r') as f:
  
    # band information
    ns = f['ns'][()]
    nbndskip = f['nbndskip'][()]
    nbnd = f['nbnd'][()]
    nocc = f['nocc'][()]
    nmode = f['nmode'][()]
  
    # k/q point information
    nQ = f['nQ'][()]
    nq = f['nq'][()]
    Qpts = f['Qpts'][()]
    qpts = f['qpts'][()]
    Q_plus_q_map = f['Q_plus_q_map'][()]
  
    # state information information
    energies = f['energies'][()]
    frequencies = f['frequencies'][()]
    gkq = f['xctph'][()]

  
  # prepare to write
  with open('xctph.dat', 'w') as f:

    for iQ, Q in enumerate(Qpts):
      for iq, q in enumerate(qpts):

        # find index for the point at k+q
        iQq = Q_plus_q_map[iQ, iq]

        # write k, q point information
        f.write('\n{:^20s}{Q[0]:8.4f}{Q[1]:8.4f}{Q[2]:8.4f}\n'
                .format('kpt = ', Q=Q))

        f.write('{:^20s}{q[0]:8.4f}{q[1]:8.4f}{q[2]:8.4f}\n\n'
                .format('qpt = ', q=q))

        f.write('{:>6s}{:>6s}{:>7s}{:>10s}{:>11s}{:>13s}{:>13s} '
                '{:>12s}{:>11s}{:>14s}\n'
                .format('nbnd', 'mbnd', 'imode',
                        'eSQ (eV)', 'eS\'Qq (eV)', 'omega (meV)',
                        'Re[G] (meV)', 'Im[G] (meV)', '|G| (meV)',
                        'Sym|G| (meV)'
                        )
                )

        eQ = energies[:, iQ] * ryd2ev
        eQq = energies[:, iQq] * ryd2ev
        freq = frequencies[:, iq] * ryd2mev
        g =  gkq[:, :, iQ, :, iq] * ryd2ev / bohr2ang  if switch_nu_and_cart_eVA else gkq[:, :, iQ, :, iq] * ryd2mev
        g_sym = sym_mtxel(g, eQ, eQq, freq)

        for nb in range(nbnd):
          for mb in range(nbnd):
            for imode in range(nmode):
              # Large write statement
              f.write(
                '{:>6d}{:>6d}{:>7d}{:>10.3f}{:>11.3f}'
                '{:>13.2f}{:>13.2f}{:>12.2f}{:>11.2f}'
                '{:>14.2f}\n'
                .format(nb + 1,
                        mb + 1,
                        imode + 1,
                        eQ[nb],
                        eQq[mb],
                        freq[imode],
                        np.real(g[mb, nb, imode]),
                        np.imag(g[mb, nb, imode]),
                        np.abs(g[mb, nb, imode]),
                        np.abs(g_sym[mb, nb, imode])
                        )
              )


if __name__ == '__main__':

  import argparse 
  parser = argparse.ArgumentParser()
  parser.add_argument('fname_xctph_h5')
  parser.add_argument('--switch_nu_and_cart_eVA', action='store_true', help='Hack to get excited state forces in eV/A.')
  args = parser.parse_args()
  
  print_xctph(**vars(args))



from setuptools import setup, find_packages

setup(
    name='xctph',
    version='1.2.0',
    description='Package for computing exciton phonon matrix elements.',
    long_description='Package for computing exciton phonon matrix elements.',
    author='Krishnaa Vadivel',
    author_email='krishnaa.vadivel@yale.edu',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    install_requires=[
        'numpy', 
        'scipy', 
        'mpi4py',
        'h5py', 
        'ase',
        'xmltodict', 
        'setuptools',
        'jmespath',
    ],
    # scripts=[     # Does not seem to work.
    #     'xctph/write_eph_h5.py',
    #     'xctph/write_xct_h5.py',
    #     'xctph/compute_xctph.py',
    #     'xctph/print_eph.py',
    #     'xctph/print_xctph.py',
    # ]
)



import numpy as np
import copy


def sym_mtxel(gkk_orig, ek, ekq, freq=None, eps=1e-3):
    """Symmetrize electron phonon matrix elements"""

    gkk = copy.deepcopy(gkk_orig)
    nbnd, nbnd, nmodes = np.shape(gkk)
    gkk_sym = abs(gkk)

    # Symmetrize over phonons first
    if freq is None:
        pass
        # print "No frequencies provided, will not symmetrize over phonon modes"
    else:
        for ibnd in range(nbnd):
            for jbnd in range(nbnd):
                for nu in range(nmodes):
                    g2 = 0
                    n = 0
                    w_1 = freq[nu]
                    for mu in range(nmodes):
                        w_2 = freq[mu]
                        if abs(w_2-w_1) < eps:
                            n = n + 1
                            g2 = g2 + abs(gkk[ibnd, jbnd, mu]) ** 2
                    g2 = g2 / float(n)
                    gkk_sym[ibnd, jbnd, nu] = np.sqrt(g2)
        gkk[:] = gkk_sym

    # Symmetrize over k+q electrons
    for nu in range(nmodes):
        for jbnd in range(nbnd):
            for ibnd in range(nbnd):
                g2 = 0
                n = 0
                e_1 = ekq[ibnd]
                for pbnd in range(nbnd):
                    e_2 = ekq[pbnd]
                    if abs(e_2-e_1) < eps:
                        n = n + 1
                        g2 = g2 + abs(gkk[pbnd, jbnd, nu]) ** 2
                g2 = g2 / float(n)
                gkk_sym[ibnd, jbnd, nu] = np.sqrt(g2)
    gkk[:] = gkk_sym

    # Symmetrize over k electrons
    for nu in range(nmodes):
        for ibnd in range(nbnd):
            for jbnd in range(nbnd):
                g2 = 0
                n = 0
                e_1 = ek[jbnd]
                for pbnd in range(nbnd):
                    e_2 = ek[pbnd]
                    if abs(e_2-e_1) < eps:
                        n = n + 1
                        g2 = g2 + abs(gkk[ibnd, pbnd, nu]) ** 2
                g2 = g2 / float(n)
                gkk_sym[ibnd, jbnd, nu] = np.sqrt(g2)
    gkk[:] = gkk_sym

    return gkk_sym
